@@ -18,9 +18,11 @@ package org.apache.pdfbox.pdfviewer;
 
 import java.awt.BasicStroke;
 import java.awt.Color;
+import java.awt.Composite;
 import java.awt.Dimension;
 import java.awt.Graphics;
 import java.awt.Graphics2D;
+import java.awt.Paint;
 import java.awt.geom.Area;
 import java.awt.RenderingHints;
 import java.awt.geom.AffineTransform;
@@ -40,6 +42,11 @@ import org.apache.pdfbox.pdmodel.common.PDRectangle;
 import org.apache.pdfbox.pdmodel.font.PDFont;
 import org.apache.pdfbox.pdmodel.graphics.PDGraphicsState;
 import org.apache.pdfbox.pdmodel.graphics.PDShading;
+import org.apache.pdfbox.pdmodel.graphics.shading.AxialShadingPaint;
+import org.apache.pdfbox.pdmodel.graphics.shading.PDShadingResources;
+import org.apache.pdfbox.pdmodel.graphics.shading.PDShadingType2;
+import org.apache.pdfbox.pdmodel.graphics.shading.PDShadingType3;
+import org.apache.pdfbox.pdmodel.graphics.shading.RadialShadingPaint;
 import org.apache.pdfbox.pdmodel.interactive.annotation.PDAnnotation;
 import org.apache.pdfbox.pdmodel.interactive.annotation.PDAppearanceDictionary;
 import org.apache.pdfbox.pdmodel.interactive.annotation.PDAppearanceStream;
@@ -49,7 +56,6 @@ import org.apache.pdfbox.util.PDFStreamEngine;
 import org.apache.pdfbox.util.ResourceLoader;
 import org.apache.pdfbox.util.TextPosition;
 import org.apache.pdfbox.cos.COSName;
-import org.apache.pdfbox.cos.COSDictionary;
 
 
 /**
@@ -64,10 +70,22 @@ public class PageDrawer extends PDFStreamEngine
     /**
      * Log instance.
      */
-    private static final Log log = LogFactory.getLog(PageDrawer.class);
+    private static final Log LOG = LogFactory.getLog(PageDrawer.class);
 
     private Graphics2D graphics;
+    
+    /**
+     * clipping winding rule used for the clipping path.
+     */
+    private int clippingWindingRule = -1;
+
+    /**
+     * Size of the page.
+     */
     protected Dimension pageSize;
+    /**
+     * Current page to be rendered.
+     */
     protected PDPage page;
 
     private GeneralPath linePath = new GeneralPath();
@@ -99,6 +117,8 @@ public class PageDrawer extends PDFStreamEngine
         pageSize = pageDimension;
         graphics.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );
         graphics.setRenderingHint( RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON );
+        // initialize the used stroke with CAP_BUTT instead of CAP_SQUARE
+        graphics.setStroke(new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER));
         // Only if there is some content, we have to process it. 
         // Otherwise we are done here and we will produce an empty page
         if ( page.getContents() != null) 
@@ -106,7 +126,7 @@ public class PageDrawer extends PDFStreamEngine
             PDResources resources = page.findResources();
             processStream( page, resources, page.getContents().getStream() );
         }
-        List annotations = page.getAnnotations();
+        List<PDAnnotation> annotations = page.getAnnotations();
         for( int i=0; i<annotations.size(); i++ )
         {
             PDAnnotation annot = (PDAnnotation)annotations.get( i );
@@ -119,15 +139,24 @@ public class PageDrawer extends PDFStreamEngine
                 {
                     appearanceName = "default";
                 }
-                Map appearanceMap = appearDictionary.getNormalAppearance();
-                if (appearanceMap != null) { 
+                Map<String, PDAppearanceStream> appearanceMap = appearDictionary.getNormalAppearance();
+                if (appearanceMap != null) 
+                { 
                     PDAppearanceStream appearance = 
                         (PDAppearanceStream)appearanceMap.get( appearanceName ); 
                     if( appearance != null ) 
                     { 
-                        g.translate( (int)rect.getLowerLeftX(), (int)-rect.getLowerLeftY() ); 
+                        Point2D point = new Point2D.Float(rect.getLowerLeftX(), rect.getLowerLeftY());
+                        Matrix matrix = appearance.getMatrix();
+                        if (matrix != null) 
+                        {
+                            // transform the rectangle using the given matrix 
+                            AffineTransform at = matrix.createAffineTransform();
+                            at.transform(point, point);
+                        }
+                        g.translate( (int)point.getX(), -(int)point.getY() );
                         processSubStream( page, appearance.getResources(), appearance.getStream() ); 
-                        g.translate( (int)-rect.getLowerLeftX(), (int)+rect.getLowerLeftY() ); 
+                        g.translate( -(int)point.getX(), (int)point.getY() ); 
                     }
                 }
             }
@@ -135,6 +164,17 @@ public class PageDrawer extends PDFStreamEngine
 
     }
 
+    /**
+     * Remove all cached resources.
+     */
+    public void dispose()
+    {
+        graphics = null;
+        linePath = null;
+        page = null;
+        pageSize = null;
+    }
+
     /**
      * You should override this method if you want to perform an action when a
      * text is being processed.
@@ -145,35 +185,48 @@ public class PageDrawer extends PDFStreamEngine
     {
         try
         {
-            switch(this.getGraphicsState().getTextState().getRenderingMode()) {
+            PDGraphicsState graphicsState = getGraphicsState();
+            Composite composite;
+            Paint paint;
+            switch(graphicsState.getTextState().getRenderingMode()) 
+            {
                 case PDTextState.RENDERING_MODE_FILL_TEXT:
-                    graphics.setComposite( this.getGraphicsState().getNonStrokeJavaComposite() );
-                    graphics.setColor( this.getGraphicsState().getNonStrokingColor().getJavaColor() );
+                    composite = graphicsState.getNonStrokeJavaComposite();
+                    paint = graphicsState.getNonStrokingColor().getJavaColor();
+                    if (paint == null)
+                    {
+                        paint = graphicsState.getNonStrokingColor().getPaint(pageSize.height);
+                    }
                     break;
                 case PDTextState.RENDERING_MODE_STROKE_TEXT:
-                    graphics.setComposite( this.getGraphicsState().getStrokeJavaComposite() );
-                    graphics.setColor( this.getGraphicsState().getStrokingColor().getJavaColor() );
+                    composite = graphicsState.getStrokeJavaComposite();
+                    paint = graphicsState.getStrokingColor().getJavaColor();
+                    if (paint == null)
+                    {
+                        paint = graphicsState.getStrokingColor().getPaint(pageSize.height);
+                    }
                     break;
                 case PDTextState.RENDERING_MODE_NEITHER_FILL_NOR_STROKE_TEXT:
                     //basic support for text rendering mode "invisible"
-                    Color nsc = this.getGraphicsState().getStrokingColor().getJavaColor();
+                    Color nsc = graphicsState.getStrokingColor().getJavaColor();
                     float[] components = {Color.black.getRed(),Color.black.getGreen(),Color.black.getBlue()};
-                    Color  c = new Color(nsc.getColorSpace(),components,0f);
-                    graphics.setComposite( this.getGraphicsState().getStrokeJavaComposite() );
-                    graphics.setColor(c);
+                    paint = new Color(nsc.getColorSpace(),components,0f);
+                    composite = graphicsState.getStrokeJavaComposite();
                     break;
                 default:
                     // TODO : need to implement....
-                    log.debug("Unsupported RenderingMode "
+                    LOG.debug("Unsupported RenderingMode "
                             + this.getGraphicsState().getTextState().getRenderingMode()
                             + " in PageDrawer.processTextPosition()."
                             + " Using RenderingMode "
                             + PDTextState.RENDERING_MODE_FILL_TEXT
                             + " instead");
-                    graphics.setComposite( this.getGraphicsState().getNonStrokeJavaComposite() );
-                    graphics.setColor( this.getGraphicsState().getNonStrokingColor().getJavaColor() );
+                    composite = graphicsState.getNonStrokeJavaComposite();
+                    paint = graphicsState.getNonStrokingColor().getJavaColor();
             }
-
+            graphics.setComposite(composite);
+            graphics.setPaint(paint);
+            
             PDFont font = text.getFont();
             Matrix textPos = text.getTextPos().copy();
             float x = textPos.getXPosition();
@@ -188,10 +241,11 @@ public class PageDrawer extends PDFStreamEngine
             AffineTransform at = textPos.createAffineTransform();
             PDMatrix fontMatrix = font.getFontMatrix();
             at.scale(fontMatrix.getValue(0, 0) * 1000f, fontMatrix.getValue(1, 1) * 1000f);
-            graphics.setClip(getGraphicsState().getCurrentClippingPath());
+            //TODO setClip() is a massive performance hot spot. Investigate optimization possibilities
+            graphics.setClip(graphicsState.getCurrentClippingPath());
             // the fontSize is no longer needed as it is already part of the transformation
             // we should remove it from the parameter list in the long run
-            font.drawString( text.getCharacter(), graphics, 1, at, x, y );
+            font.drawString( text.getCharacter(), text.getCodePoints(), graphics, 1, at, x, y );
         }
         catch( IOException io )
         {
@@ -278,7 +332,18 @@ public class PageDrawer extends PDFStreamEngine
     public void fillPath(int windingRule) throws IOException
     {
         graphics.setComposite(getGraphicsState().getNonStrokeJavaComposite());
-        graphics.setColor( getGraphicsState().getNonStrokingColor().getJavaColor() );
+        Paint nonStrokingPaint = getGraphicsState().getNonStrokingColor().getJavaColor();
+        if ( nonStrokingPaint == null )
+        {
+            nonStrokingPaint = getGraphicsState().getNonStrokingColor().getPaint(pageSize.height);
+        }
+        if ( nonStrokingPaint == null )
+        {
+            LOG.info("ColorSpace "+getGraphicsState().getNonStrokingColor().getColorSpace().getName()
+                    +" doesn't provide a non-stroking color, using white instead!");
+            nonStrokingPaint = Color.WHITE;
+        }
+        graphics.setPaint( nonStrokingPaint );
         getLinePath().setWindingRule(windingRule);
         graphics.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF );
         graphics.setClip(getGraphicsState().getCurrentClippingPath());
@@ -317,7 +382,18 @@ public class PageDrawer extends PDFStreamEngine
     public void strokePath() throws IOException
     {
         graphics.setComposite(getGraphicsState().getStrokeJavaComposite());
-        graphics.setColor( getGraphicsState().getStrokingColor().getJavaColor() );
+        Paint strokingPaint = getGraphicsState().getStrokingColor().getJavaColor();
+        if ( strokingPaint == null )
+        {
+            strokingPaint = getGraphicsState().getStrokingColor().getPaint(pageSize.height);
+        }
+        if ( strokingPaint == null )
+        {
+            LOG.info("ColorSpace "+getGraphicsState().getStrokingColor().getColorSpace().getName()
+                    +" doesn't provide a stroking color, using white instead!");
+            strokingPaint = Color.WHITE;
+        }
+        graphics.setPaint(strokingPaint);
         graphics.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF );
         graphics.setClip(getGraphicsState().getCurrentClippingPath());
         GeneralPath path = getLinePath();
@@ -357,27 +433,52 @@ public class PageDrawer extends PDFStreamEngine
      *
      * @param windingRule The winding rule this path will use.
      * 
+     * @deprecated use {@link #setClippingWindingRule(int)} instead
+     * 
      */
     public void setClippingPath(int windingRule)
     {
-        PDGraphicsState graphicsState = getGraphicsState();
-        GeneralPath clippingPath = (GeneralPath)getLinePath().clone();
-        clippingPath.setWindingRule(windingRule);
-        // If there is already set a clipping path, we have to intersect the new with the existing one
-        if (graphicsState.getCurrentClippingPath() != null) 
+        setClippingWindingRule(windingRule);
+    }
+    
+    /**
+     * Set the clipping winding rule.
+     *
+     * @param windingRule The winding rule which will be used for clipping.
+     * 
+     */
+    public void setClippingWindingRule(int windingRule)
+    {
+        clippingWindingRule = windingRule;
+    }
+
+    /**
+     * Set the clipping Path.
+     *
+     */
+    public void endPath()
+    {
+        if (clippingWindingRule > -1)
         {
-            Area currentArea = new Area(getGraphicsState().getCurrentClippingPath());
-            Area newArea = new Area(clippingPath);
-            currentArea.intersect(newArea);
-            graphicsState.setCurrentClippingPath(currentArea);
-        }
-        else 
-        {
-            graphicsState.setCurrentClippingPath(clippingPath);
+            PDGraphicsState graphicsState = getGraphicsState();
+            GeneralPath clippingPath = (GeneralPath)getLinePath().clone();
+            clippingPath.setWindingRule(clippingWindingRule);
+            // If there is already set a clipping path, we have to intersect the new with the existing one
+            if (graphicsState.getCurrentClippingPath() != null) 
+            {
+                Area currentArea = new Area(getGraphicsState().getCurrentClippingPath());
+                Area newArea = new Area(clippingPath);
+                currentArea.intersect(newArea);
+                graphicsState.setCurrentClippingPath(currentArea);
+            }
+            else 
+            {
+                graphicsState.setCurrentClippingPath(clippingPath);
+            }
+            clippingWindingRule = -1;
         }
         getLinePath().reset();
     }
-    
     /**
      * Draw the AWT image. Called by Invoke.
      * Moved into PageDrawer so that Invoke doesn't have to reach in here for Graphics as that breaks extensibility.
@@ -386,7 +487,8 @@ public class PageDrawer extends PDFStreamEngine
      * @param at The transformation to use when drawing.
      * 
      */
-    public void drawImage(Image awtImage, AffineTransform at){
+    public void drawImage(Image awtImage, AffineTransform at)
+    {
         graphics.setComposite(getGraphicsState().getStrokeJavaComposite());
         graphics.setClip(getGraphicsState().getCurrentClippingPath());
         graphics.drawImage( awtImage, at, null );
@@ -398,61 +500,55 @@ public class PageDrawer extends PDFStreamEngine
      * @param ShadingName  The name of the Shading Dictionary to use for this fill instruction.
      *
      * @throws IOException If there is an IO error while shade-filling the path/clipping area.
+     * 
+     * @deprecated use {@link #shFill(COSName)) instead.
      */
     public void SHFill(COSName ShadingName) throws IOException
     {
-        PDShading Shading =FindShadingDictionary(ShadingName);
-        log.info("Shading = " + Shading.toString());
-        
-        switch (Shading.getShadingType()){
-            case 1:
-                SHFill_Function(Shading);
-            break;
-            case 2:
-                SHFill_Axial(Shading);
-                break;
-            case 3:
-                SHFill_Radial(Shading);
-                break;
-            case 4:
-                SHFill_FreeGourad(Shading);
-                break;
-            case 5:
-                SHFill_LatticeGourad(Shading);
-                break;
-            case 6:
-                SHFill_CoonsPatch(Shading);
-                break;
-            case 7:
-                SHFill_TensorPatch(Shading);
-                break;
-            
-            default:
-                throw new IOException("Invalid ShadingType " + Shading.getShadingType() + " for Shading " + ShadingName);
-        }
+        shFill(ShadingName);
     }
     
     /**
-     * Find the appropriate Shading Dictionary. This is its own private function as it is really not appropriate to override when deriving from PageDrawer.
+     * Fill with Shading.  Called by SHFill operator.
      *
-     * @param ShadingName  The name of the Shading Dictionary to use for this fill instruction.
+     * @param shadingName  The name of the Shading Dictionary to use for this fill instruction.
      *
-     * @returns The PDShading object
-     * @throws IOException If there is an IO error while attempting to find the appropriate PDShading object.
+     * @throws IOException If there is an IO error while shade-filling the clipping area.
      */
-    private PDShading FindShadingDictionary(COSName ShadingName) throws IOException
+    public void shFill(COSName shadingName) throws IOException
     {
-        
-        PDResources resources = (PDResources)page.getResources();
-                
-        COSDictionary AllShadings = (COSDictionary)(resources.getCOSDictionary().getDictionaryObject(COSName.SHADING));
-        
-        PDShading Shading = new PDShading(ShadingName, (COSDictionary)(AllShadings.getDictionaryObject(ShadingName)));
-        
-        return Shading;
-        
+        PDShadingResources shading = getResources().getShadings().get(shadingName.getName());
+        LOG.debug("Shading = " + shading.toString());
+        int shadingType = shading.getShadingType();
+        Matrix ctm = getGraphicsState().getCurrentTransformationMatrix();
+        Paint paint = null;
+        switch (shadingType)
+        {
+            case 1:
+                // TODO
+                LOG.debug("Function based shading not yet supported");
+            break;
+            case 2:
+                paint = new AxialShadingPaint((PDShadingType2)shading, ctm, pageSize.height);
+                break;
+            case 3:
+                paint = new RadialShadingPaint((PDShadingType3)shading, ctm, pageSize.height);
+                break;
+            case 4:
+            case 5:
+            case 6:
+            case 7:
+                // TODO
+                LOG.debug("Shading type "+shadingType+" not yet supported");
+                break;
+            default:
+                throw new IOException("Invalid ShadingType " + shadingType + " for Shading " + shadingName);
+        }
+        graphics.setComposite(getGraphicsState().getNonStrokeJavaComposite());
+        graphics.setPaint(paint);
+        graphics.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF );
+        graphics.fill( getGraphicsState().getCurrentClippingPath() );
     }
-    
     /**
      * Fill with a Function-based gradient / shading.  
      * If extending the class, override this and its siblings, not the public SHFill method.
